import shutil
from pathlib import Path

import pytest

from zettings.utils import get_nested_value, is_valid_key, set_nested_value


@pytest.fixture
def temp_filepath(tmpdir_factory: pytest.TempdirFactory):
    temp_dir = str(tmpdir_factory.mktemp("temp"))
    temp_testing_dir = Path(temp_dir) / "testing"
    temp_testing_dir.mkdir(parents=True, exist_ok=True)
    yield Path(temp_testing_dir)
    shutil.rmtree(temp_dir)


@pytest.mark.parametrize(
    ("key", "expected"),
    [
        ("DEBUG", True),
        ("debug-mode", True),
        ("db.host", False),
        ('"db.host"', False),
        ("db host", False),
        ("'db host'", False),
        ('"bad"key"', False),
        ("'bad'key'", False),
        ("simple_key", True),
        ("123key", True),
        ("key-with-dash", True),
        ("key with space", False),
        ('"key with space in nested quotes"', False),
        ('"key with \n newline"', False),
        ("'single'quote'", False),
        ("'valid key'", False),
        ('"valid key"', False),
        ('""', False),
        ("", False),
    ],
)
def test_is_valid_key(key, expected):
    assert is_valid_key(key) == expected


def test_get_nested_invalid_key():
    d = {"a": {"b": {"c": 1}}}
    with pytest.raises(ValueError):  # noqa: PT011
        get_nested_value(d, "a.b ")
    with pytest.raises(ValueError):  # noqa: PT011
        get_nested_value(d, "b ")
    with pytest.raises(ValueError):  # noqa: PT011
        get_nested_value(d, "b .c.d.a")


def test_set_nested_invalid_key():
    d = {"a": {"b": {"c": 1}}}
    with pytest.raises(ValueError):
        set_nested_value(d, "a.b .c", 2)
    with pytest.raises(ValueError):
        set_nested_value(d, "b ", 2)
    with pytest.raises(ValueError):
        set_nested_value(d, "b .c.d.a", 2)
    with pytest.raises(ValueError):
        set_nested_value(d, "a.b.c ", 2)


@pytest.mark.parametrize(
    "initial, key, value, expected",
    [
        ({}, "a.b.c", 1, {"a": {"b": {"c": 1}}}),
        ({"a": {}}, "a.b", 2, {"a": {"b": 2}}),
        ({"a": {"b": 3}}, "a.b", 4, {"a": {"b": 4}}),
        ({}, "x", 5, {"x": 5}),
        ({"a": {"b": {"c": 1}}}, "a.b.d", 6, {"a": {"b": {"c": 1, "d": 6}}}),
    ],
)
def test_set_nested(initial, key, value, expected):
    d = initial.copy()
    set_nested_value(d, key, value)
    assert d == expected


@pytest.mark.parametrize(
    "d, key, expected",
    [
        ({"a": {"b": {"c": 1}}}, "a.b.c", 1),
        ({"a": {"b": 2}}, "a.b", 2),
        ({"x": 5}, "x", 5),
        ({"a": {"b": {"c": 1}}}, "a.b", {"c": 1}),
        ({"a": {"b": {"c": 1}}}, "a.b.c.d", None),
        ({}, "foo.bar", None),
        ({"a": 1}, "a.b", None),
    ],
)
def test_get_nested(d, key, expected):
    assert get_nested_value(d, key) == expected


def test_set_nested_with_custom_separator():
    d = {}
    set_nested_value(d, "a|b|c", 10, sep="|")
    assert d == {"a": {"b": {"c": 10}}}


def test_get_nested_with_custom_separator():
    d = {"a": {"b": {"c": 42}}}
    assert get_nested_value(d, "a|b|c", sep="|") == 42


def test_set_nested_cannot_overwrite_non_dict():
    d = {"a": 1}

    with pytest.raises(KeyError):
        set_nested_value(d, "a.b", 3)

    with pytest.raises(KeyError):
        set_nested_value(d, "a.b.c", 3)
    assert d == {"a": 1}


def test_get_nested_returns_none_for_non_dict():
    d = {"a": 1}
    assert get_nested_value(d, "a.b") is None



import shutil
from pathlib import Path

import pytest

from zettings import Settings

default_settings_normal_format = {
    "settings": {"name": "MyName", "mood": "MyMood"},
    "dictionary": {
        "key1": "value1",
        "key2": "value2",
        "subdictionary": {"key1": "subvalue1", "key2": "subvalue2"},
    },
}
default_settings_nested_format = {
    "settings.name": "MyName",
    "settings.mood": "MyMood",
    "dictionary.key1": "value1",
    "dictionary.key2": "value2",
    "dictionary.subdictionary.key1": "subvalue1",
    "dictionary.subdictionary.key2": "subvalue2",
}
default_settings_invalid_normal_format = {
    "settings": {"name": "MyName", "mood": "MyMood"},
    "dictionary": {
        "key1'invalid'": "value1",
        "key2": "value2",
        "subdictionary": {"key1": "subvalue1", "key2": "subvalue2"},
    },
}
default_settings_invalid_nested_format = {
    "settings.name": "MyName",
    "settings.mood": "MyMood",
    "dictionary.key1": "value1",
    "dictionary.key2": "value2",
    "dictionary.subdict'invalidionary.key1": "subvalue1",
    "dictionary.subdictionary.key2": "subvalue2",
}


@pytest.fixture
def settings_filepath(tmpdir_factory: pytest.TempdirFactory):
    temp_dir = str(tmpdir_factory.mktemp("temp"))
    temp_testing_dir = temp_dir + "/testing/settings.toml"
    yield Path(temp_testing_dir)
    shutil.rmtree(temp_dir)


@pytest.fixture
def temp_home():
    name = ".test-blah-blah-test-only/settings.toml"
    path = Path.home() / name
    temp_dir = path.parent
    if path.exists():
        path.unlink()

    temp_home = {}
    temp_home["dir"] = temp_dir
    temp_home["name"] = name
    temp_home["path"] = path
    yield temp_home

    if path.exists():
        path.unlink()

    if temp_dir.exists():
        temp_dir.rmdir()


def test_settings_initializes_with_empty_file(settings_filepath):
    settings = Settings(settings_filepath)
    assert settings.get("settings.name") is None
    assert settings.get("settings.mood") is None
    assert settings.get("dictionary.key1") is None
    assert settings.get("dictionary") is None

    # Check if metadata keys are present
    assert settings.get("metadata.notice") is not None
    assert settings.get("metadata.created") is not None
    assert settings.get("metadata.updated") is not None
    assert settings.get("metadata.NotReal") is None


def test_settings_initializes_with_default_settings_normal_format(settings_filepath):
    settings = Settings(settings_filepath, defaults=default_settings_normal_format)
    for k, v in default_settings_normal_format.items():
        assert settings.get(k) == v

    assert settings.get("settings.name") == "MyName"
    assert settings.get("settings.mood") == "MyMood"
    assert settings.get("dictionary.key1") == "value1"
    assert settings.get("dictionary") == {
        "key1": "value1",
        "key2": "value2",
        "subdictionary": {"key1": "subvalue1", "key2": "subvalue2"},
    }

    assert settings.get("foo") is None


def test_settings_initializes_with_default_settings_nested_format(settings_filepath):
    settings = Settings(settings_filepath, defaults=default_settings_nested_format)
    for k, v in default_settings_nested_format.items():
        assert settings.get(k) == v

    assert settings.get("settings.name") == "MyName"
    assert settings.get("settings.mood") == "MyMood"
    assert settings.get("dictionary.key1") == "value1"
    assert settings.get("dictionary") == {
        "key1": "value1",
        "key2": "value2",
        "subdictionary": {"key1": "subvalue1", "key2": "subvalue2"},
    }

    assert settings.get("foo") is None


## Type Tests
@pytest.mark.parametrize(
    ("value"),
    [
        (523),
        (None),
        (["settings.toml"]),
        ({"settings.toml": "value"}),
    ],
)
def test_settings_name_type_fails(value):
    with pytest.raises(TypeError):
        _ = Settings(value, read_only=True)


@pytest.mark.parametrize(
    ("value"),
    [
        "blah spaces/settings.toml",
        "blah spaces and dashes-/settings.toml",
        "/n/settings.toml",
        "./settings.toml",
        ";/settings.toml",
        "",
        "/settings.toml",
        "test/.toml",
        ".test/settings.json",
    ],
)
def test_settings_name_alphanumericish_fails(value):
    with pytest.raises(ValueError):  # noqa: PT011
        Settings(value, read_only=True)


@pytest.mark.parametrize(
    ("value"),
    [
        "blah_spaces/settings.toml",
        "123/settings.toml",
        "abc/settings3.toml",
        "ABC/config.toml",
        "ABCabc123_-/settings.TOML",
        ".blah-blah/settings.toml",
        r"blah-blah\settings.toml",
    ],
)
def test_settings_name_alphanumericish_passes(value):
    _ = Settings(value, read_only=True)


@pytest.mark.parametrize(
    ("value"),
    [
        (523),
        ("abc"),
    ],
)
def test_settings_defaults_type_fails(settings_filepath, value):
    with pytest.raises(TypeError):
        _ = Settings(filepath=settings_filepath, defaults=value)


@pytest.mark.parametrize(
    ("value"),
    [
        (523),
        ("abc"),
        (["settings.toml"]),
        ({"settings.toml": "value"}),
    ],
)
def test_settings_always_reloads_type_fails(settings_filepath, value):
    with pytest.raises(TypeError):
        _ = Settings(filepath=settings_filepath, defaults=value, always_reload=value)


@pytest.mark.parametrize(
    ("value"),
    [
        (523),
        ("abc"),
        (["settings.toml"]),
        ({"settings.toml": "value"}),
    ],
)
def test_settings_read_only_type_fails(settings_filepath, value):
    with pytest.raises(TypeError):
        _ = Settings(filepath=settings_filepath, defaults=value, read_only=value)


@pytest.mark.parametrize(
    ("value"),
    [
        (523),
        (False),
        (["settings.toml"]),
        ({"settings.toml": "value"}),
    ],
)
def test_settings_filepath_type_fails(value):
    with pytest.raises(TypeError):
        _ = Settings(filepath=value)


## End Type Tests
def test_settings_sets_missing_keys_in_defaults(settings_filepath):
    defaults = default_settings_normal_format.copy()
    settings = Settings(filepath=settings_filepath, defaults=defaults)
    assert settings.get("foo") is None

    defaults["foo"] = "bar"
    new_settings = Settings(filepath=settings_filepath, defaults=defaults)

    assert new_settings.get("foo") == "bar"
    for k, v in defaults.items():
        assert new_settings.get(k) == v


def test_settings_get_and_set_methods_success(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)

    settings.set("string", "string_value")
    settings.set("none", None)
    settings.set("int", 42)
    settings.set("float", 3.14)
    settings.set("bool", True)  # noqa: FBT003
    settings.set("list", [1, 2, 3])
    settings.set("dict", {"key": "value"})
    settings.set("nested_dict", {"nested_key": {"sub_key": "sub_value"}})
    settings.set("empty_list", [])
    settings.set("nested_list", [[1, 2, 3], [4, 5, 6], ["a", "b", "c"]])
    settings.set("empty_dict", {})
    settings.set("complex", {"list": [1, 2, 3], "dict": {"key": "value"}})
    settings.set("complex_nested", {"outer": {"inner": {"key": "value"}}})
    settings.set("unicode", "こんにちは")  # Japanese for "Hello"
    settings.set("emoji", "😊")  # Smiling face emoji

    assert settings.get("string") == "string_value"
    assert settings.get("none") is None
    assert settings.get("int") == 42
    assert settings.get("float") == 3.14
    assert settings.get("bool") is True
    assert settings.get("list") == [1, 2, 3]
    assert settings.get("dict") == {"key": "value"}
    assert settings.get("nested_dict") == {"nested_key": {"sub_key": "sub_value"}}
    assert settings.get("empty_list") == []
    assert settings.get("nested_list") == [[1, 2, 3], [4, 5, 6], ["a", "b", "c"]]
    assert settings.get("empty_dict") == {}
    assert settings.get("complex") == {"list": [1, 2, 3], "dict": {"key": "value"}}
    assert settings.get("complex_nested") == {"outer": {"inner": {"key": "value"}}}
    assert settings.get("unicode") == "こんにちは"
    assert settings.get("emoji") == "😊"


def test_settings_overrides_existing_settings(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)

    # Set an initial value
    settings.set("name", "InitialName")
    assert settings.get("name") == "InitialName"

    # Override the value
    settings.set("name", "NewName")
    assert settings.get("name") == "NewName"


def test_settings_handles_non_existent_keys(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)

    assert settings.get("non_existent_key") is None


def test_settings_handles_empty_settings_file(settings_filepath):
    # Create an empty settings file
    settings_filepath.parent.mkdir(parents=True, exist_ok=True)
    with Path.open(settings_filepath, "w") as f:
        f.write("")
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)
    # Check that default settings are applied
    for k, v in default_settings_normal_format.items():
        assert settings.get(k) == v


def test_settings_handles_creating_directories_for_new_files(settings_filepath):
    parent_dir = settings_filepath.parent

    assert not parent_dir.exists(), "Parent directory should not exist before test"

    _ = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)
    assert parent_dir.exists(), "Parent directory should be created by Settings class"


def test_settings_saves_settings_to_file(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)

    # Set some values
    settings.set("name", "TestName")
    settings.set("mood", "TestMood")

    # Reload the settings to check if values are saved
    new_settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)
    assert new_settings.get("name") == "TestName"
    assert new_settings.get("mood") == "TestMood"


def test_settings_with_different_cases_in_key(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)
    settings["caseCheck"] = "value"

    assert settings["caseCheck"] == "value"
    assert settings["casecheck"] is None


def test_settings_no_default_settings(settings_filepath):
    # Test with no default settings
    settings = Settings(filepath=settings_filepath)

    # Check that no settings are set initially
    assert settings.get("name") is None
    assert settings.get("mood") is None

    # Set a value and check if it persists
    settings.set("name", "NoDefaultName")
    assert settings.get("name") == "NoDefaultName"

    # Reload the settings to check if the value is saved
    new_settings = Settings(filepath=settings_filepath)
    assert new_settings.get("name") == "NoDefaultName"


def test_settings_with_getitem_and_setitem(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)

    # Test __getitem__
    assert settings["settings.name"] == "MyName"
    assert settings["settings.mood"] == "MyMood"
    assert settings["dictionary.key1"] == "value1"

    # Test __setitem__
    settings["settings.name"] = "NewName"
    assert settings["settings.name"] == "NewName"

    settings["new_key"] = "new_value"
    assert settings["new_key"] == "new_value"


def test_settings_updates_defaults_with_nested_dict(settings_filepath: Path):
    defaults = default_settings_nested_format.copy()
    settings = Settings(filepath=settings_filepath, defaults=defaults)
    assert settings.get("dictionary.subdictionary.key3") is None
    assert settings["dictionary.subdictionary.key3"] is None

    defaults["dictionary.subdictionary.key3"] = "subvalue3"

    # Access a nested value
    new_settings = Settings(filepath=settings_filepath, defaults=defaults)
    assert new_settings.get("dictionary.subdictionary.key3") == "subvalue3"
    assert new_settings["dictionary"]["subdictionary"]["key3"] == "subvalue3"


def test_settings_initializes_defaults_with_nested_dict(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_nested_format)
    assert settings.get("settings.name") == "MyName"
    assert settings.get("settings.mood") == "MyMood"
    assert settings["settings"]["name"] == "MyName"
    assert settings["settings"]["mood"] == "MyMood"
    assert settings.get("dictionary") is not None
    assert settings.get("dictionary.key1") == "value1"
    assert settings.get("dictionary.subdictionary") is not None
    assert settings.get("dictionary.subdictionary.key1") == "subvalue1"


def test_settings_sets_default_settings_of_nested_dictionaries_if_not_present(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_nested_format)
    assert settings.get("settings.mood") == "MyMood"
    assert settings.get("settings.face") is None

    new_default_settings = default_settings_nested_format.copy()
    new_default_settings["settings.face"] = "round"

    new_settings = Settings(filepath=settings_filepath, defaults=new_default_settings)

    assert settings.get("settings.face") == "round"
    assert new_settings.get("settings.face") == "round"


def test_settings_always_reload_true(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)

    assert settings.get("settings.name") == "MyName"

    # Change the settings file seperatrely
    settings2 = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)
    settings2.set("settings.name", "NewName")

    # Verify that the change is reflected in the original settings object
    assert settings.get("settings.name") == "NewName"


def test_settings_always_reload_false(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)
    settings.always_reload = False

    assert settings.get("settings.name") == "MyName"

    # Change the settings file seperatrely
    settings2 = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)
    settings2.set("settings.name", "NewName")


def test_settings_dynamic_reload_true_set(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)

    assert settings.get("settings.name") == "MyName"

    # Change the settings file seperatrely
    settings2 = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)
    settings2.set("settings.name", "NewName3")

    # Verify that the change is reflected in the original settings object
    settings2.set("settings.mood", "MyMood")
    assert settings.get("settings.name") == "NewName3"


def test_settings_dynamic_reload_false_set(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)
    settings.always_reload = False

    assert settings.get("settings.name") == "MyName"

    # Change the settings file seperatrely
    settings2 = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)
    settings2.set("settings.name", "NewName3")

    # Verify that the change is reflected in the original settings object
    settings2.set("settings.mood", "MyMood")
    assert settings.get("settings.name") == "MyName"

    settings.set("settings.name", "NewName4")
    assert settings.get("settings.name") == "NewName4"


def test_settings_read_only_true(settings_filepath):
    settings = Settings(
        filepath=settings_filepath,
        defaults=default_settings_normal_format,
        read_only=True,
    )

    assert settings.read_only is True
    with pytest.raises(PermissionError):
        settings.set("settings.name", "NewName")


# Verify that the change is reflected in the original settings object
def test_settings_repr_returns_expected_string(settings_filepath):
    settings = Settings(
        filepath=settings_filepath,
        defaults=default_settings_normal_format,
    )
    expected = f"Settings stored at: {settings_filepath}"
    assert repr(settings) == expected


def test_settings_stores_in_home_directory_if_no_filepath(temp_home):
    settings = Settings(temp_home["name"])
    assert temp_home["path"].exists() is True


def test_settings_fails_with_invalid_defaults_format(settings_filepath):
    with pytest.raises(ValueError):
        Settings(filepath=settings_filepath, defaults=default_settings_invalid_normal_format)

    # Verify that the settings file is not created
    assert not settings_filepath.exists()

    with pytest.raises(ValueError):
        Settings(filepath=settings_filepath, defaults=default_settings_invalid_nested_format)

    # Verify that the settings file is not created
    assert not settings_filepath.exists()


def test_settings_iter_and_len_method_with_reload_true(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format, always_reload=True)
    settings2 = Settings(filepath=settings_filepath, defaults=default_settings_normal_format, always_reload=False)

    settings2.set("newkey", "newkeyvalue")

    # Test the iter method
    keys = list(settings)
    assert "settings" in keys
    assert "metadata" in keys
    assert "dictionary" in keys
    assert "newkey" in keys

    assert len(keys) == 4


def test_settings_iter_and_len_method_with_reload_false(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format, always_reload=False)
    settings2 = Settings(filepath=settings_filepath, defaults=default_settings_normal_format, always_reload=False)

    settings2.set("newkey", "newkeyvalue")

    keys = list(settings)
    assert "settings" in keys
    assert "metadata" in keys
    assert "dictionary" in keys
    assert "newkey" not in keys

    assert len(keys) == 3


def test_settings_del_method(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)

    settings.set("newkey", "newkeyvalue")
    assert settings.get("newkey") == "newkeyvalue"
    assert settings["newkey"] is not None
    del settings["newkey"]

    assert settings.get("newkey") is None
    assert settings["newkey"] is None

    # Test Nested Key Deletion
    settings.set("nested.key", "nested_value")
    settings.set("nested.another_key", "another_value")
    assert settings.get("nested.key") == "nested_value"
    assert settings.get("nested.another_key") == "another_value"

    del settings["nested.key"]
    assert settings.get("nested.key") is None
    assert settings.get("nested.another_key") == "another_value"

    settings.set("nested.key", "nested_value")
    settings.set("nested.another_key", "another_value")
    assert settings.get("nested.key") == "nested_value"
    assert settings.get("nested.another_key") == "another_value"
    del settings["nested.another_key"]
    assert settings.get("nested.key") == "nested_value"
    assert settings.get("nested.another_key") is None


def test_settings_delete_with_read_only_true(settings_filepath):
    settings = Settings(filepath=settings_filepath, defaults=default_settings_normal_format)

    settings.read_only = True
    with pytest.raises(PermissionError):
        del settings["settings.name"]

    # Verify that the settings file is not modified
    assert settings.get("settings.name") == "MyName"
    assert settings["settings.name"] == "MyName"


def test_get_with_duplicate_keynames(settings_filepath):
    defaults = {
        "key1.test": "value1",
        "key1.subkey.subsubkey": "value2",
        "key2.test": "value3",
        "key2.subkey.subsubkey": "value4",
        "key3.test": "value5",
        "key3.subkey.subsubkey": "value6",
    }

    settings = Settings(filepath=settings_filepath, defaults=defaults)

    assert settings.get("key1.test") == "value1"
    assert settings.get("key1.subkey.subsubkey") == "value2"
    assert settings.get("key2.test") == "value3"
    assert settings.get("key2.subkey.subsubkey") == "value4"
    assert settings.get("key3.test") == "value5"
    assert settings.get("key3.subkey.subsubkey") == "value6"

    settings.set("key1.test", "new_value1")
    settings.set("key1.subkey.subsubkey", "new_value2")
    settings.set("key2.test", "new_value3")
    settings.set("key2.subkey.subsubkey", "new_value4")
    settings.set("key3.test", "new_value5")
    settings.set("key3.subkey.subsubkey", "new_value6")

    assert settings.get("key1.test") == "new_value1"
    assert settings.get("key1.subkey.subsubkey") == "new_value2"
    assert settings.get("key2.test") == "new_value3"
    assert settings.get("key2.subkey.subsubkey") == "new_value4"
    assert settings.get("key3.test") == "new_value5"
    assert settings.get("key3.subkey.subsubkey") == "new_value6"


def test_another_invalid_defaults_format(settings_filepath):
    defaults = {
        "key1.subkey": "value1",
        "key1.subkey.subsubkey": "value2",
    }

    with pytest.raises(KeyError):
        _ = Settings(filepath=settings_filepath, defaults=defaults)



from __future__ import annotations

import re
from collections.abc import MutableMapping
from datetime import datetime, timezone
from pathlib import Path
from threading import Lock
from typing import Any

import toml

from zettings.utils import delete_nested_key, get_nested_value, set_nested_value, validate_dictionary

# Constants for metadata
NOTICE_KEY = "metadata.notice"
NOTICE = "This file was created by zettings."
CREATED_KEY = "metadata.created"
UPDATED_KEY = "metadata.updated"


class Settings(MutableMapping[str, Any]):
    lock = Lock()

    def __init__(
        self,
        filepath: str | Path,
        defaults: dict | None = None,
        *,
        always_reload: bool = True,
        read_only: bool = False,
    ) -> None:
        """Initialize the settings manager.

        Args:
            filepath (str): Must be in the format '.name/subname.toml'
                name and subname must be alphanumeric, underscores, or dashes.
                the '.' in front of name is optional.
                Examples: 'myproject/settings.toml', '.myproject/settings.toml'.

                When using a str for filepath, the Path object will be created using Path.home() / filepath.
            filepath (Path): An explicit path to the settings file.

            defaults (dict | None, optional): A dictionary of default settings. Defaults to None.
            always_reload (bool, optional): Whether to always reload settings from file. Defaults to True.
            read_only (bool, optional): If True, disables writing to the settings file. Defaults to False.


        Raises:
            TypeError: If any argument is of an incorrect type.
            ValueError: If filepath (str) does not match the required format.

        """
        # Validate the arguments
        if not isinstance(filepath, (str, Path)):
            error_message = "filepath must be a string or Path"
            raise TypeError(error_message)
        if isinstance(filepath, str):
            if not re.fullmatch(r"^(?:\.)?[a-zA-Z0-9_-]+[\\/]{1}[a-zA-Z0-9_-]+\.toml$", filepath, re.IGNORECASE):
                error_message = (
                    "home_filepath must be alphanumeric, underscores, or dashes in format '.name/subname.toml'"
                )
                raise ValueError(error_message)
            self._filepath = Path.home() / filepath
        else:
            self._filepath = filepath  # assume Path object
        if not isinstance(defaults, (dict, type(None))):
            error_message = "defaults must be a dictionary or None"
            raise TypeError(error_message)
        if not isinstance(always_reload, bool):
            error_message = "always_reload must be a boolean"
            raise TypeError(error_message)
        if not isinstance(read_only, bool):
            error_message = "read_only must be a boolean"
            raise TypeError(error_message)

        self.always_reload = always_reload
        self.read_only = read_only
        self._data = {}
        if defaults is None:
            defaults = {}

        validate_dictionary(defaults)
        if not self.read_only:
            self._initialize_file()
            self._initialize_defaults(defaults)

    def _initialize_file(self) -> None:
        """Initialize the settings file with metadata values and sets defaults for missing keys."""
        if self._filepath.exists():
            return
        self._filepath.parent.mkdir(parents=True, exist_ok=True)
        set_nested_value(self._data, NOTICE_KEY, NOTICE)
        set_nested_value(self._data, CREATED_KEY, datetime.now(tz=timezone.utc).isoformat())
        set_nested_value(self._data, UPDATED_KEY, datetime.now(tz=timezone.utc).isoformat())
        self._save()

    def _initialize_defaults(self, d: dict, parent_key: str = "") -> None:
        """Recursively set default values for missing keys in the settings dictionary."""
        for k, v in d.items():
            full_key = f"{parent_key}.{k}" if parent_key else k
            if self.get(full_key) is None:
                self.set(full_key, v)
            elif isinstance(v, dict):
                self._initialize_defaults(v, full_key)

    def _save(self) -> None:
        """Save the settings to the file and update the updated timestamp."""
        set_nested_value(self._data, UPDATED_KEY, datetime.now(tz=timezone.utc).isoformat())
        with self.lock, Path.open(self._filepath, mode="w", encoding="utf-8") as f:
            toml.dump(self._data, f)

    def _load(self) -> None:
        """Load the settings from the file."""
        with self.lock, Path.open(self._filepath, mode="r", encoding="utf-8") as f:
            self._data = toml.load(f)

    def get(self, key: str) -> Any | None:  # noqa: ANN401
        """Return a value from the configuration by key.

        Args:
        key (str): The key to return a value from.

        Returns:
        Any | None: The value associated with the key, or None if the key does not exist.

        """
        if self.always_reload:
            self._load()
        return get_nested_value(self._data, key)

    def set(self, key: str, value: Any) -> None:  # noqa: ANN401
        """Set a value in the configuration by key.

        Args:
        key (str): The key to store the value under.
        value (Any): The value to set for the key.

        Returns:
        None

        """
        if self.read_only:
            error_message = "Settings are read-only and cannot be modified."
            raise PermissionError(error_message)

        if self.always_reload:
            self._load()

        set_nested_value(self._data, key, value)
        self._save()

    def __getitem__(self, key: str) -> Any | None:  # noqa: ANN401
        """Get an item from the configuration by key."""
        return self.get(key)

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: ANN401
        """Set an item in the configuration by key."""
        self.set(key, value)

    def __repr__(self) -> str:  # noqa: D105
        return f"Settings stored at: {self._filepath}"

    def __delitem__(self, key: str) -> None:
        """Delete an item from the settings by key.

        Args:
        key (str): The key to delete from the settings file.

        Returns:
        None

        Raises:
        PermissionError: If the settings are read only.

        """
        if self.read_only:
            error_message = "Settings are read only and cannot be modified."
            raise PermissionError(error_message)

        if self.always_reload:
            self._load()

        delete_nested_key(self._data, key)
        self._save()

    def __iter__(self):
        """Return an iterator over the keys in the settings."""
        if self.always_reload:
            self._load()
        return iter(self._data)

    def __len__(self) -> int:
        """Return the number of items in the settings."""
        if self.always_reload:
            self._load()
        return len(self._data)
